#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from hopsworks import client


class VariableApi:
    def __init__(self):
        pass

    def get_variable(self, variable: str):
        """Get the configured value for a variable"""

        _client = client.get_instance()

        path_params = ["variables", variable]
        domain = _client._send_request("GET", path_params)

        return domain["successMessage"]

    def get_version(self, software: str):
        _client = client.get_instance()
        path_params = [
            "variables",
            "versions",
        ]

        resp = _client._send_request("GET", path_params)
        for entry in resp:
            if entry["software"] == software:
                return entry["version"]
        return None



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import math
import os
import time
from tqdm.auto import tqdm
import shutil
import logging
import copy

from hopsworks import client
from hopsworks.client.exceptions import RestAPIError
from hopsworks.client.exceptions import DatasetException
from concurrent.futures import ThreadPoolExecutor, wait


class Chunk:
    def __init__(self, content, number, status):
        self.content = content
        self.number = number
        self.status = status
        self.retries = 0


class DatasetApi:
    def __init__(
        self,
        project_id,
    ):
        self._project_id = project_id
        self._log = logging.getLogger(__name__)

    DEFAULT_FLOW_CHUNK_SIZE = 1048576
    FLOW_PERMANENT_ERRORS = [404, 413, 415, 500, 501]

    def download(self, path: str, local_path: str = None, overwrite: bool = False):
        """Download file from Hopsworks Filesystem to the current working directory.

        ```python

        import hopsworks

        project = hopsworks.login()

        dataset_api = project.get_dataset_api()

        downloaded_file_path = dataset_api.download("Resources/my_local_file.txt")

        ```
        # Arguments
            path: path in Hopsworks filesystem to the file
            local_path: path where to download the file in the local filesystem
            overwrite: overwrite local file if exists
        # Returns
            `str`: Path to downloaded file
        # Raises
            `RestAPIError`: If unable to download the file
        """
        _client = client.get_instance()
        path_params = [
            "project",
            self._project_id,
            "dataset",
            "download",
            "with_auth",
            path,
        ]
        query_params = {"type": "DATASET"}

        # Build the path to download the file on the local fs and return to the user, it should be absolute for consistency
        # Download in CWD if local_path not specified
        if local_path is None:
            local_path = os.path.join(os.getcwd(), os.path.basename(path))
        # If local_path specified, ensure it is absolute
        else:
            if os.path.isabs(local_path):
                local_path = os.path.join(local_path, os.path.basename(path))
            else:
                local_path = os.path.join(
                    os.getcwd(), local_path, os.path.basename(path)
                )

        if os.path.exists(local_path):
            if overwrite:
                if os.path.isfile:
                    os.remove(local_path)
                else:
                    shutil.rmtree(local_path)
            else:
                raise IOError(
                    "{} already exists, set overwrite=True to overwrite it".format(
                        local_path
                    )
                )

        file_size = int(self._get(path)["attributes"]["size"])
        with _client._send_request(
            "GET", path_params, query_params=query_params, stream=True
        ) as response:
            with open(local_path, "wb") as f:
                pbar = None
                try:
                    pbar = tqdm(
                        total=file_size,
                        bar_format="{desc}: {percentage:.3f}%|{bar}| {n_fmt}/{total_fmt} elapsed<{elapsed} remaining<{remaining}",
                        desc="Downloading",
                    )
                except Exception:
                    self._log.exception("Failed to initialize progress bar.")
                    self._log.info("Starting download")

                for chunk in response.iter_content(
                    chunk_size=self.DEFAULT_FLOW_CHUNK_SIZE
                ):
                    f.write(chunk)

                    if pbar is not None:
                        pbar.update(len(chunk))

                if pbar is not None:
                    pbar.close()
                else:
                    self._log.info("Download finished")

        return local_path

    def upload(
        self,
        local_path: str,
        upload_path: str,
        overwrite: bool = False,
        chunk_size=1048576,
        simultaneous_uploads=3,
        max_chunk_retries=1,
        chunk_retry_interval=1,
    ):
        """Upload a file to the Hopsworks filesystem.

        ```python

        import hopsworks

        project = hopsworks.login()

        dataset_api = project.get_dataset_api()

        uploaded_file_path = dataset_api.upload("my_local_file.txt", "Resources")

        ```
        # Arguments
            local_path: local path to file to upload
            upload_path: path to directory where to upload the file in Hopsworks Filesystem
            overwrite: overwrite file if exists
            chunk_size: upload chunk size in bytes. Default 1048576 bytes
            simultaneous_uploads: number of simultaneous chunks to upload. Default 3
            max_chunk_retries: maximum retry for a chunk. Default is 1
            chunk_retry_interval: chunk retry interval in seconds. Default is 1sec
        # Returns
            `str`: Path to uploaded file
        # Raises
            `RestAPIError`: If unable to upload the file
        """
        # local path could be absolute or relative,
        if not os.path.isabs(local_path) and os.path.exists(
            os.path.join(os.getcwd(), local_path)
        ):
            local_path = os.path.join(os.getcwd(), local_path)

        file_size = os.path.getsize(local_path)

        _, file_name = os.path.split(local_path)

        destination_path = upload_path + "/" + file_name

        if self.exists(destination_path):
            if overwrite:
                self.remove(destination_path)
            else:
                raise DatasetException(
                    "{} already exists, set overwrite=True to overwrite it".format(
                        local_path
                    )
                )

        num_chunks = math.ceil(file_size / chunk_size)

        base_params = self._get_flow_base_params(
            file_name, num_chunks, file_size, chunk_size
        )

        chunk_number = 1
        with open(local_path, "rb") as f:
            pbar = None
            try:
                pbar = tqdm(
                    total=file_size,
                    bar_format="{desc}: {percentage:.3f}%|{bar}| {n_fmt}/{total_fmt} elapsed<{elapsed} remaining<{remaining}",
                    desc="Uploading",
                )
            except Exception:
                self._log.exception("Failed to initialize progress bar.")
                self._log.info("Starting upload")
            with ThreadPoolExecutor(simultaneous_uploads) as executor:
                while True:
                    chunks = []
                    for _ in range(simultaneous_uploads):
                        chunk = f.read(chunk_size)
                        if not chunk:
                            break
                        chunks.append(Chunk(chunk, chunk_number, "pending"))
                        chunk_number += 1

                    if len(chunks) == 0:
                        break

                    # upload each chunk and update pbar
                    futures = [
                        executor.submit(
                            self._upload_chunk,
                            base_params,
                            upload_path,
                            file_name,
                            chunk,
                            pbar,
                            max_chunk_retries,
                            chunk_retry_interval,
                        )
                        for chunk in chunks
                    ]
                    # wait for all upload tasks to complete
                    _, _ = wait(futures)
                    try:
                        _ = [future.result() for future in futures]
                    except Exception as e:
                        if pbar is not None:
                            pbar.close()
                        raise e

            if pbar is not None:
                pbar.close()
            else:
                self._log.info("Upload finished")

        return upload_path + "/" + os.path.basename(local_path)

    def _upload_chunk(
        self,
        base_params,
        upload_path,
        file_name,
        chunk: Chunk,
        pbar,
        max_chunk_retries,
        chunk_retry_interval,
    ):
        query_params = copy.copy(base_params)
        query_params["flowCurrentChunkSize"] = len(chunk.content)
        query_params["flowChunkNumber"] = chunk.number

        chunk.status = "uploading"
        while True:
            try:
                self._upload_request(
                    query_params, upload_path, file_name, chunk.content
                )
                break
            except RestAPIError as re:
                chunk.retries += 1
                if (
                    re.response.status_code in DatasetApi.FLOW_PERMANENT_ERRORS
                    or chunk.retries > max_chunk_retries
                ):
                    chunk.status = "failed"
                    raise re
                time.sleep(chunk_retry_interval)
                continue

        chunk.status = "uploaded"

        if pbar is not None:
            pbar.update(query_params["flowCurrentChunkSize"])

    def _get_flow_base_params(self, file_name, num_chunks, size, chunk_size):
        return {
            "templateId": -1,
            "flowChunkSize": chunk_size,
            "flowTotalSize": size,
            "flowIdentifier": str(size) + "_" + file_name,
            "flowFilename": file_name,
            "flowRelativePath": file_name,
            "flowTotalChunks": num_chunks,
        }

    def _upload_request(self, params, path, file_name, chunk):
        _client = client.get_instance()
        path_params = ["project", self._project_id, "dataset", "upload", path]

        # Flow configuration params are sent as form data
        _client._send_request(
            "POST", path_params, data=params, files={"file": (file_name, chunk)}
        )

    def _get(self, path: str):
        """Get dataset.

        :param path: path to check
        :type path: str
        :return: dataset metadata
        :rtype: dict
        """
        _client = client.get_instance()
        path_params = ["project", self._project_id, "dataset", path]
        headers = {"content-type": "application/json"}
        return _client._send_request("GET", path_params, headers=headers)

    def exists(self, path: str):
        """Check if a file exists in the Hopsworks Filesystem.

        # Arguments
            path: path to check
        # Returns
            `bool`: True if exists, otherwise False
        # Raises
            `RestAPIError`: If unable to check existence for the path
        """
        try:
            self._get(path)
            return True
        except RestAPIError:
            return False

    def remove(self, path: str):
        """Remove a path in the Hopsworks Filesystem.

        # Arguments
            path: path to remove
        # Raises
            `RestAPIError`: If unable to remove the path
        """
        _client = client.get_instance()
        path_params = ["project", self._project_id, "dataset", path]
        _client._send_request("DELETE", path_params)

    def mkdir(self, path: str):
        """Create a directory in the Hopsworks Filesystem.

        ```python

        import hopsworks

        project = hopsworks.login()

        dataset_api = project.get_dataset_api()

        directory_path = dataset_api.mkdir("Resources/my_dir")

        ```
        # Arguments
            path: path to directory
        # Returns
            `str`: Path to created directory
        # Raises
            `RestAPIError`: If unable to create the directory
        """
        _client = client.get_instance()
        path_params = ["project", self._project_id, "dataset", path]
        query_params = {
            "action": "create",
            "searchable": "true",
            "generate_readme": "false",
            "type": "DATASET",
        }
        headers = {"content-type": "application/json"}
        return _client._send_request(
            "POST", path_params, headers=headers, query_params=query_params
        )["attributes"]["path"]

    def copy(self, source_path: str, destination_path: str, overwrite: bool = False):
        """Copy a file or directory in the Hopsworks Filesystem.

        ```python

        import hopsworks

        project = hopsworks.login()

        dataset_api = project.get_dataset_api()

        directory_path = dataset_api.copy("Resources/myfile.txt", "Logs/myfile.txt")

        ```
        # Arguments
            source_path: the source path to copy
            destination_path: the destination path
            overwrite: overwrite destination if exists
        # Raises
            `RestAPIError`: If unable to perform the copy
        """
        if self.exists(destination_path):
            if overwrite:
                self.remove(destination_path)
            else:
                raise DatasetException(
                    "{} already exists, set overwrite=True to overwrite it".format(
                        destination_path
                    )
                )

        _client = client.get_instance()
        path_params = ["project", self._project_id, "dataset", source_path]
        query_params = {
            "action": "copy",
            "destination_path": destination_path,
        }
        _client._send_request("POST", path_params, query_params=query_params)

    def move(self, source_path: str, destination_path: str, overwrite: bool = False):
        """Move a file or directory in the Hopsworks Filesystem.

        ```python

        import hopsworks

        project = hopsworks.login()

        dataset_api = project.get_dataset_api()

        directory_path = dataset_api.move("Resources/myfile.txt", "Logs/myfile.txt")

        ```
        # Arguments
            source_path: the source path to move
            destination_path: the destination path
            overwrite: overwrite destination if exists
        # Raises
            `RestAPIError`: If unable to perform the move
        """

        if self.exists(destination_path):
            if overwrite:
                self.remove(destination_path)
            else:
                raise DatasetException(
                    "{} already exists, set overwrite=True to overwrite it".format(
                        destination_path
                    )
                )

        _client = client.get_instance()
        path_params = ["project", self._project_id, "dataset", source_path]
        query_params = {
            "action": "move",
            "destination_path": destination_path,
        }
        _client._send_request("POST", path_params, query_params=query_params)



#
#   Copyright 2023 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import time
from hopsworks.engine import execution_engine
from hopsworks.core import execution_api
from hopsworks.core import flink_cluster_api
from hopsworks import util


class FlinkCluster:
    def __init__(
        self,
        job,
        project_id=None,
        project_name=None,
        **kwargs,
    ):
        self._job = job
        self._project_id = project_id
        self._project_name = project_name
        self._execution_id = None

        self._execution_engine = execution_engine.ExecutionEngine(project_id)
        self._execution_api = execution_api.ExecutionsApi(project_id)
        self._flink_cluster_api = flink_cluster_api.FlinkClusterApi(
            project_id, project_name
        )

    def _get_execution(self, assert_running=False):
        current_execution = None
        if self._job is not None:
            executions = self._job.get_executions()
            for exec in executions:
                # If previous start was used on the cluster, always refer to the specific execution_id
                if self._execution_id is not None:
                    if exec.id == self._execution_id:
                        current_execution = exec
                        break
                # Otherwise get the first running one and set the id
                elif exec.success is None:
                    self._execution_id = exec.id
                    current_execution = exec
                    break
        if assert_running is True and current_execution.success is not None:
            raise Exception(
                "This operation requires the FlinkCluster to be running. To start it invoke .start()"
            )
        return current_execution

    def _count_ongoing_executions(self):
        ongoing_executions = 0
        if self._job is not None:
            executions = self._job.get_executions()
            for exec in executions:
                if exec.success is None:
                    ongoing_executions += 1
        return ongoing_executions

    def start(self, await_time=1800):
        """Start the flink cluster and wait until it reaches RUNNING state.

        ```python

        import hopsworks

        project = hopsworks.login()

        flink_cluster_api = project.get_flink_cluster_api()

        flink_cluster = flink_cluster_api.get_cluster(name="myFlinkCluster")

        flink_cluster.start()
        ```
        # Arguments
            await_time: defaults to 1800 seconds to account for auto-scale mechanisms.
        # Raises
            `RestAPIError`: If unable to start the flink cluster.
        """

        if self._count_ongoing_executions() > 0:
            raise Exception(
                "There is already a running FlinkCluster. Use FlinkClusterApi.get_cluster('{}') to get a reference to it.".format(
                    self._job.name
                )
            )

        execution = self._execution_api._start(self._job)
        updated_execution = self._execution_api._get(self._job, execution.id)
        while updated_execution.state == "INITIALIZING":
            updated_execution = self._execution_api._get(self._job, execution.id)
            if updated_execution.state == "RUNNING":
                print("Cluster is running")
                break

            self._execution_engine._log.info(
                "Waiting for cluster to start. Current state: {}.".format(
                    updated_execution.state
                )
            )

            await_time -= 1
            time.sleep(3)

        if updated_execution.state != "RUNNING":
            raise Exception(
                "FlinkCluster {} did not start within the allocated time and exited with state {}".format(
                    execution.id, execution.state
                )
            )

        self._execution_id = execution.id
        return self

    def get_jobs(self):
        """Get jobs from the flink cluster.
        ```python

        # log in to hopsworks
        import hopsworks
        project = hopsworks.login()

        # fetch flink cluster handle
        flink_cluster_api = project.get_flink_cluster_api()
        flink_cluster = flink_cluster_api.get_cluster(name="myFlinkCluster")

        # get jobs from this flink cluster
        flink_cluster.get_jobs()
        ```

        # Returns
            `List[Dict]`: The array of dicts with flink job id and status of the job.
        # Raises
            `RestAPIError`: If unable to get the jobs from the cluster
        """

        return self._flink_cluster_api._get_jobs(
            self._get_execution(assert_running=True)
        )

    def get_job(self, job_id):
        """Get specific job from the flink cluster.
        ```python

        # log in to hopsworks
        import hopsworks
        project = hopsworks.login()

        # fetch flink cluster handle
        flink_cluster_api = project.get_flink_cluster_api()
        flink_cluster = flink_cluster_api.get_cluster(name="myFlinkCluster")

        # get jobs from this cluster
        job_id = '113a2af5b724a9b92085dc2d9245e1d6'
        flink_cluster.get_job(job_id)
        ```

        # Arguments
            job_id: id of the job within this cluster
        # Returns
            `Dict`: Dict with flink job id and status of the job.
        # Raises
            `RestAPIError`: If unable to get the job from the cluster
        """

        return self._flink_cluster_api._get_job(
            self._get_execution(assert_running=True), job_id
        )

    def stop_job(self, job_id):
        """Stop specific job in the flink cluster.
        ```python

        # log in to hopsworks
        import hopsworks
        project = hopsworks.login()

        # fetch flink cluster handle
        flink_cluster_api = project.get_flink_cluster_api()
        flink_cluster = flink_cluster_api.get_cluster(name="myFlinkCluster")

        # stop the job
        job_id = '113a2af5b724a9b92085dc2d9245e1d6'
        flink_cluster.stop_job(job_id)
        ```

        # Arguments
            job_id: id of the job within this flink cluster.
        # Raises
            `RestAPIError`: If unable to stop the job
        """
        self._flink_cluster_api._stop_job(
            self._get_execution(assert_running=True), job_id
        )

    def get_jars(self):
        """Get already uploaded jars from the flink cluster.
        ```python
        # log in to hopsworks
        import hopsworks
        project = hopsworks.login()

        # fetch flink cluster handle
        flink_cluster_api = project.get_flink_cluster_api()
        flink_cluster = flink_cluster_api.get_cluster(name="myFlinkCluster")

        # get jar files from this cluster
        flink_cluster.get_jars()
        ```

        # Returns
            `List[Dict]`: The array of dicts with jar metadata.
        # Raises
            `RestAPIError`: If unable to get jars from the flink cluster.
        """
        return self._flink_cluster_api._get_jars(
            self._get_execution(assert_running=True)
        )

    def upload_jar(self, jar_file):
        """Upload jar file to the flink cluster.
        ```python
        # log in to hopsworks
        import hopsworks
        project = hopsworks.login()

        # fetch flink cluster handle
        flink_cluster_api = project.get_flink_cluster_api()
        flink_cluster = flink_cluster_api.get_cluster(name="myFlinkCluster")

        # upload jar file to this cluster
        jar_file_path = "./flink-example.jar"
        flink_cluster.upload_jar(jar_file_path)
        ```

        # Arguments
            jar_file: path to the jar file.
        # Raises
            `RestAPIError`: If unable to upload jar file
        """

        self._flink_cluster_api._upload_jar(
            self._get_execution(assert_running=True), jar_file
        )

    def submit_job(self, jar_id, main_class, job_arguments=None):
        """Submit job using the specific jar file uploaded to the flink cluster.
        ```python
        # log in to hopsworks
        import hopsworks
        project = hopsworks.login()

        # fetch flink cluster handle
        flink_cluster_api = project.get_flink_cluster_api()
        flink_cluster = flink_cluster_api.get_cluster(name="myFlinkCluster")

        # upload jar file to this cluster
        main_class = "com.example.Main"
        job_arguments = "-arg1 arg1 -arg2 arg2"
        jar_file_path = "./flink-example.jar"
        flink_cluster.upload_jar(jar_file_path)

        #get jar file metadata (and select the 1st one for demo purposes)
        jar_metadata = flink_cluster.get_jars()[0]
        jar_id = jar_metadata["id"]
        flink_cluster.submit_job(jar_id, main_class, job_arguments=job_arguments)
        ```

        # Arguments
            jar_id: id if the jar file
            main_class: path to the main class of the jar file
            job_arguments: Job arguments (if any), defaults to none.
        # Returns
            `str`:  job id.
        # Raises
            `RestAPIError`: If unable to submit the job.
        """

        return self._flink_cluster_api._submit_job(
            self._get_execution(assert_running=True), jar_id, main_class, job_arguments
        )

    def job_state(self, job_id):
        """Gets state of the job submitted to the flink cluster.
        ```python

        # log in to hopsworks
        import hopsworks
        project = hopsworks.login()

        # fetch flink cluster handle
        flink_cluster_api = project.get_flink_cluster_api()
        flink_cluster = flink_cluster_api.get_cluster(name="myFlinkCluster")

        # get jobs from this flink cluster
        job_id = '113a2af5b724a9b92085dc2d9245e1d6'
        flink_cluster.job_state(job_id)
        ```

        # Arguments
            job_id: id of the job within this flink cluster
        # Returns
            `str`: status of the job. Possible states:  "INITIALIZING", "CREATED", "RUNNING", "FAILING", "FAILED",
            "CANCELLING", "CANCELED",  "FINISHED", "RESTARTING", "SUSPENDED", "RECONCILING".
        # Raises
            `RestAPIError`: If unable to get the job state from the flink cluster.
        """

        return self._flink_cluster_api._job_state(
            self._get_execution(assert_running=True), job_id
        )

    def stop(self):
        """Stop this cluster.
        ```python

        # log in to hopsworks
        import hopsworks
        project = hopsworks.login()

        # fetch flink cluster handle
        flink_cluster_api = project.get_flink_cluster_api()

        flink_cluster = flink_cluster_api.get_cluster(name="myFlinkCluster")

        flink_cluster.stop()
        ```

        # Raises
            `RestAPIError`: If unable to stop the flink cluster.
        """
        exec = self._get_execution(assert_running=False)
        if exec is None or exec.success is not None:
            return
        else:
            self._flink_cluster_api._stop_execution(
                self._get_execution(assert_running=False)
            )

    @property
    def id(self):
        """Id of the cluster"""
        return self._job._id

    @property
    def name(self):
        """Name of the cluster"""
        return self._job._name

    @property
    def creation_time(self):
        """Date of creation for the cluster"""
        return self._job._creation_time

    @property
    def config(self):
        """Configuration for the cluster"""
        return self._job._config

    @property
    def creator(self):
        """Creator of the cluster"""
        return self._job._creator

    @property
    def state(self):
        """State of the cluster"""
        current_exec = self._get_execution(assert_running=False)
        if current_exec is not None:
            return current_exec.state
        else:
            return None

    def get_url(self):
        path = "/p/" + str(self._project_id) + "/jobs/named/" + self.name
        return (
            "FlinkCluster created successfully, explore it at "
            + util.get_hostname_replaced_url(path)
        )



#
#   Copyright 2023 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from unittest import TestCase, mock
from contextlib import contextmanager
import hopsworks
import getpass
import os
import uuid
import tempfile
import importlib
import shutil
from hopsworks.client import exceptions
from hopsworks.project import Project


@contextmanager
def input(*cmds):
    """Replace input."""
    hidden_cmds = [c.get("hidden") for c in cmds if isinstance(c, dict)]
    with mock.patch("getpass.getpass", side_effect=hidden_cmds):
        yield


class TestLogin(TestCase):
    """Test hopsworks login."""

    def setUp(self):
        if not hasattr(self, "system_tmp"):
            importlib.reload(tempfile)
            self.system_tmp = tempfile.gettempdir()

        self.cwd_path = os.path.join(self.system_tmp, str(uuid.uuid4()))
        self.temp_dir = os.path.join(self.cwd_path, "tmp")
        self.home_dir = os.path.join(self.cwd_path, "home")

        os.environ["TMP"] = self.temp_dir
        os.environ["HOME"] = self.home_dir

        os.mkdir(self.cwd_path)
        os.mkdir(self.temp_dir)
        os.mkdir(self.home_dir)

        importlib.reload(tempfile)

        os.chdir(self.cwd_path)

    def tearDown(self):
        os.environ["TMP"] = self.system_tmp
        shutil.rmtree(self.cwd_path)
        hopsworks.logout()

    def _check_api_key_existence(self):

        path = hopsworks._get_cached_api_key_path()

        api_key_name = ".hw_api_key"
        api_key_folder = ".{}_hopsworks_app".format(getpass.getuser())

        # Path for current working directory api key
        cwd_api_key_path = os.path.join(os.getcwd(), api_key_name)

        # Path for home api key
        home_dir_path = os.path.expanduser("~")
        home_api_key_path = os.path.join(home_dir_path, api_key_folder, api_key_name)

        # Path for tmp api key
        temp_dir_path = tempfile.gettempdir()
        temp_api_key_path = os.path.join(temp_dir_path, api_key_folder, api_key_name)

        return (
            path,
            path == cwd_api_key_path,
            path == home_api_key_path,
            path == temp_api_key_path,
        )

    def test_login_api_key_as_input(self):
        # Should accept api key as input from command line

        path, in_cwd, in_home, in_tmp = self._check_api_key_existence()

        assert in_cwd is False
        assert in_home is True and not os.path.exists(path)
        assert in_tmp is False

        with input({"hidden": os.environ["APP_API_KEY"]}):
            project = hopsworks.login()
            self.assertTrue(isinstance(project, Project))

        path, in_cwd, in_home, in_tmp = self._check_api_key_existence()

        assert in_cwd is False
        assert in_home is True and os.path.exists(path)
        assert in_tmp is False

    def test_login_api_key_as_argument(self):
        # Should accept api key as argument
        path, in_cwd, in_home, in_tmp = self._check_api_key_existence()

        assert in_cwd is False
        assert in_home is True and not os.path.exists(path)
        assert in_tmp is False
        # Should create API key in home by default
        project = hopsworks.login(api_key_value=os.environ["APP_API_KEY"])
        self.assertTrue(isinstance(project, Project))

        self.api_key_path, in_cwd, in_home, in_tmp = self._check_api_key_existence()

        assert in_cwd is False
        assert in_home is True and not os.path.exists(path)
        assert in_tmp is False

    def test_login_cmd_input_incorrect(self):
        # Should fail to login with incorrect API key

        path, in_cwd, in_home, in_tmp = self._check_api_key_existence()

        assert in_cwd is False
        assert in_home is True and not os.path.exists(path)
        assert in_tmp is False

        with self.assertRaises(exceptions.RestAPIError):
            with input({"hidden": "incorrect_api_key"}):
                hopsworks.login()

    def test_login_fallback_to_tmp(self):
        # Should fall back to storing api key in tmp folder if home is not write and executable for user
        os.chmod(self.home_dir, 0o400)

        path, in_cwd, in_home, in_tmp = self._check_api_key_existence()

        assert in_cwd is False
        assert in_home is False
        assert in_tmp is True and not os.path.exists(path)

        # Should use API key in tmp folder
        with input({"hidden": os.environ["APP_API_KEY"]}):
            project = hopsworks.login()
            self.assertTrue(isinstance(project, Project))

        assert in_cwd is False
        assert in_home is False
        assert in_tmp is True and os.path.exists(path)

    def test_login_use_cwd_api_key(self):
        # Should use API key in cwd if exists

        api_key_path = os.path.join(os.getcwd(), ".hw_api_key")
        with open(api_key_path, "w") as f:
            f.write(os.environ["APP_API_KEY"])

        path, in_cwd, in_home, in_tmp = self._check_api_key_existence()

        assert in_cwd is True and os.path.exists(path)
        assert in_home is False
        assert in_tmp is False

        project = hopsworks.login()
        self.assertTrue(isinstance(project, Project))

        path, in_cwd, in_home, in_tmp = self._check_api_key_existence()

        assert in_cwd is True and os.path.exists(path)
        assert in_home is False
        assert in_tmp is False

    def test_login_use_home_api_key(self):
        # Should use API key in home if exists

        api_key_folder_path = os.path.join(
            os.path.expanduser("~"), ".{}_hopsworks_app".format(getpass.getuser())
        )
        os.mkdir(api_key_folder_path)
        with open(os.path.join(api_key_folder_path, ".hw_api_key"), "w") as f:
            f.write(os.environ["APP_API_KEY"])

        path, in_cwd, in_home, in_tmp = self._check_api_key_existence()

        assert in_cwd is False
        assert in_home is True and os.path.exists(path)
        assert in_tmp is False

        project = hopsworks.login()
        self.assertTrue(isinstance(project, Project))

        path, in_cwd, in_home, in_tmp = self._check_api_key_existence()

        assert in_cwd is False
        assert in_home is True and os.path.exists(path)
        assert in_tmp is False

    def test_login_api_key_as_environ(self):
        # Should accept api key as environmet variable
        try:
            os.environ["HOPSWORKS_API_KEY"] = os.environ["APP_API_KEY"]

            path, in_cwd, in_home, in_tmp = self._check_api_key_existence()

            assert in_cwd is False
            assert in_home is True and not os.path.exists(path)
            assert in_tmp is False

            project = hopsworks.login()
            self.assertTrue(isinstance(project, Project))

            path, in_cwd, in_home, in_tmp = self._check_api_key_existence()

            assert in_cwd is False
            assert in_home is True and not os.path.exists(path)
            assert in_tmp is False
        except Exception as e:
            raise e
        finally:
            del os.environ["HOPSWORKS_API_KEY"]

    def test_login_newline_in_api_key(self):
        try:
            imaginaryApiKey = "ImaginaryApiKey\n"
            hopsworks.login(api_key_value=imaginaryApiKey)
        except Exception as e:
            self.assertNotIn(imaginaryApiKey.strip(), str(e))



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import humps
import json
from hopsworks.engine import execution_engine
from hopsworks.core import execution_api
from hopsworks import constants, util


class Execution:
    def __init__(
        self,
        id=None,
        state=None,
        final_status=None,
        submission_time=None,
        stdout_path=None,
        stderr_path=None,
        app_id=None,
        hdfs_user=None,
        args=None,
        progress=None,
        user=None,
        files_to_remove=None,
        duration=None,
        flink_master_url=None,
        monitoring=None,
        type=None,
        href=None,
        project_id=None,
        job=None,
        **kwargs,
    ):
        self._id = id
        self._final_status = final_status
        self._state = state
        self._submission_time = submission_time
        self._stdout_path = stdout_path
        self._stderr_path = stderr_path
        self._args = args
        self._progress = progress
        self._user = user
        self._duration = duration
        self._monitoring = monitoring
        self._app_id = app_id
        self._hdfs_user = hdfs_user
        self._job = job
        self._project_id = project_id

        self._execution_engine = execution_engine.ExecutionEngine(project_id)
        self._execution_api = execution_api.ExecutionsApi(project_id)

    @classmethod
    def from_response_json(cls, json_dict, project_id, job):
        json_decamelized = humps.decamelize(json_dict)
        if "count" not in json_decamelized:
            return cls(**json_decamelized, project_id=project_id, job=job)
        elif json_decamelized["count"] == 0:
            return []
        else:
            return [
                cls(**execution, project_id=project_id, job=job)
                for execution in json_decamelized["items"]
            ]

    def update_from_response_json(self, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        self.__init__(**json_decamelized)
        return self

    @property
    def id(self):
        """Id of the execution"""
        return self._id

    @property
    def job_name(self):
        """Name of the job the execution belongs to"""
        return self._job.name

    @property
    def job_type(self):
        """Type of the job the execution belongs to"""
        return self._job.job_type

    @property
    def state(self):
        """Current state of the execution."""
        return self._state

    @property
    def final_status(self):
        """Final status of the execution. Can be UNDEFINED, SUCCEEDED, FAILED or KILLED."""
        return self._final_status

    @property
    def submission_time(self):
        """Timestamp when the execution was submitted"""
        return self._submission_time

    @property
    def stdout_path(self):
        """Path in Hopsworks Filesystem to stdout log file"""
        return self._stdout_path

    @property
    def stderr_path(self):
        """Path in Hopsworks Filesystem to stderr log file"""
        return self._stderr_path

    @property
    def app_id(self):
        """Application id for the execution"""
        return self._app_id

    @property
    def hdfs_user(self):
        """Filesystem user for the execution."""
        return self._hdfs_user

    @property
    def args(self):
        """Arguments set for the execution."""
        return self._args

    @property
    def progress(self):
        """Progress of the execution."""
        return self._progress

    @property
    def user(self):
        """User that submitted the execution."""
        return self._user

    @property
    def duration(self):
        """Duration in milliseconds the execution ran."""
        return self._duration

    @property
    def success(self):
        """Boolean to indicate if execution ran successfully or failed

        # Returns
            `bool`. True if execution ran successfully. False if execution failed or was killed.
        """

        is_yarn_job = (
            self.job_type.lower() == "spark"
            or self.job_type.lower() == "pyspark"
            or self.job_type.lower() == "flink"
        )

        if is_yarn_job:
            if self.final_status in constants.JOBS.ERROR_STATES:
                return False
            elif self.final_status in constants.JOBS.SUCCESS_STATES:
                return True
        else:
            if self.state in constants.JOBS.ERROR_STATES:
                return False
            elif self.state in constants.JOBS.SUCCESS_STATES:
                return True
        return None

    def download_logs(self, path=None):
        """Download stdout and stderr logs for the execution
        Example for downloading and printing the logs

        ```python

        # Download logs
        out_log_path, err_log_path = execution.download_logs()

        out_fd = open(out_log_path, "r")
        print(out_fd.read())

        err_fd = open(err_log_path, "r")
        print(err_fd.read())

        ```

        # Arguments
            path: path to download the logs. must be `str`
        # Returns
            `str`. Path to downloaded log for stdout.
            `str`. Path to downloaded log for stderr.
        """
        return self._execution_engine.download_logs(self, path)

    def delete(self):
        """Delete the execution
        !!! danger "Potentially dangerous operation"
            This operation deletes the execution.
        # Raises
            `RestAPIError`.
        """
        self._execution_api._delete(self._job.name, self.id)

    def await_termination(self):
        """Wait until execution reaches terminal state
        # Raises
            `RestAPIError`.
        """
        self._execution_engine.wait_until_finished(self._job, self)

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def __str__(self):
        return self.json()

    def __repr__(self):
        return f"Execution({self._final_status!r}, {self._state!r}, {self._submission_time!r}, {self._args!r})"

    def get_url(self):
        path = (
            "/p/"
            + str(self._project_id)
            + "/jobs/named/"
            + self.job_name
            + "/executions"
        )
        return "Execution started, explore it at " + util.get_hostname_replaced_url(
            path
        )



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import base64
import os
import textwrap
from pathlib import Path

import requests

from hopsworks.client import base, auth

try:
    import jks
except ImportError:
    pass


class Client(base.Client):
    REQUESTS_VERIFY = "REQUESTS_VERIFY"
    DOMAIN_CA_TRUSTSTORE_PEM = "DOMAIN_CA_TRUSTSTORE_PEM"
    PROJECT_ID = "HOPSWORKS_PROJECT_ID"
    PROJECT_NAME = "HOPSWORKS_PROJECT_NAME"
    HADOOP_USER_NAME = "HADOOP_USER_NAME"
    MATERIAL_DIRECTORY = "MATERIAL_DIRECTORY"
    HDFS_USER = "HDFS_USER"
    T_CERTIFICATE = "t_certificate"
    K_CERTIFICATE = "k_certificate"
    TRUSTSTORE_SUFFIX = "__tstore.jks"
    KEYSTORE_SUFFIX = "__kstore.jks"
    PEM_CA_CHAIN = "ca_chain.pem"
    CERT_KEY_SUFFIX = "__cert.key"
    MATERIAL_PWD = "material_passwd"
    SECRETS_DIR = "SECRETS_DIR"

    def __init__(self):
        """Initializes a client being run from a job/notebook directly on Hopsworks."""
        self._base_url = self._get_hopsworks_rest_endpoint()
        self._host, self._port = self._get_host_port_pair()
        self._secrets_dir = (
            os.environ[self.SECRETS_DIR] if self.SECRETS_DIR in os.environ else ""
        )
        self._cert_key = self._get_cert_pw()
        trust_store_path = self._get_trust_store_path()
        hostname_verification = (
            os.environ[self.REQUESTS_VERIFY]
            if self.REQUESTS_VERIFY in os.environ
            else "true"
        )
        self._project_id = os.environ[self.PROJECT_ID]
        self._project_name = self._project_name()
        try:
            self._auth = auth.BearerAuth(self._read_jwt())
        except FileNotFoundError:
            self._auth = auth.ApiKeyAuth(self._read_apikey())
        self._verify = self._get_verify(hostname_verification, trust_store_path)
        self._session = requests.session()

        self._connected = True

        credentials = self._get_credentials(self._project_id)

        self._write_pem_file(
            credentials["caChain"], self._get_ca_chain_path(self._project_name)
        )
        self._write_pem_file(
            credentials["clientCert"], self._get_client_cert_path(self._project_name)
        )
        self._write_pem_file(
            credentials["clientKey"], self._get_client_key_path(self._project_name)
        )

    def _get_hopsworks_rest_endpoint(self):
        """Get the hopsworks REST endpoint for making requests to the REST API."""
        return os.environ[self.REST_ENDPOINT]

    def _get_trust_store_path(self):
        """Convert truststore from jks to pem and return the location"""
        ca_chain_path = Path(self.PEM_CA_CHAIN)
        if not ca_chain_path.exists():
            self._write_ca_chain(ca_chain_path)
        return str(ca_chain_path)

    def _get_ca_chain_path(self, project_name) -> str:
        return os.path.join("/tmp", "ca_chain.pem")

    def _get_client_cert_path(self, project_name) -> str:
        return os.path.join("/tmp", "client_cert.pem")

    def _get_client_key_path(self, project_name) -> str:
        return os.path.join("/tmp", "client_key.pem")

    def _write_ca_chain(self, ca_chain_path):
        """
        Converts JKS trustore file into PEM to be compatible with Python libraries
        """
        keystore_pw = self._cert_key
        keystore_ca_cert = self._convert_jks_to_pem(
            self._get_jks_key_store_path(), keystore_pw
        )
        truststore_ca_cert = self._convert_jks_to_pem(
            self._get_jks_trust_store_path(), keystore_pw
        )

        with ca_chain_path.open("w") as f:
            f.write(keystore_ca_cert + truststore_ca_cert)

    def _convert_jks_to_pem(self, jks_path, keystore_pw):
        """
        Converts a keystore JKS that contains client private key,
         client certificate and CA certificate that was used to
         sign the certificate to PEM format and returns the CA certificate.
        Args:
        :jks_path: path to the JKS file
        :pw: password for decrypting the JKS file
        Returns:
             strings: (ca_cert)
        """
        # load the keystore and decrypt it with password
        ks = jks.KeyStore.load(jks_path, keystore_pw, try_decrypt_keys=True)
        ca_certs = ""

        # Convert CA Certificates into PEM format and append to string
        for alias, c in ks.certs.items():
            ca_certs = ca_certs + self._bytes_to_pem_str(c.cert, "CERTIFICATE")
        return ca_certs

    def _bytes_to_pem_str(self, der_bytes, pem_type):
        """
        Utility function for creating PEM files

        Args:
            der_bytes: DER encoded bytes
            pem_type: type of PEM, e.g Certificate, Private key, or RSA private key

        Returns:
            PEM String for a DER-encoded certificate or private key
        """
        pem_str = ""
        pem_str = pem_str + "-----BEGIN {}-----".format(pem_type) + "\n"
        pem_str = (
            pem_str
            + "\r\n".join(
                textwrap.wrap(base64.b64encode(der_bytes).decode("ascii"), 64)
            )
            + "\n"
        )
        pem_str = pem_str + "-----END {}-----".format(pem_type) + "\n"
        return pem_str

    def _get_jks_trust_store_path(self):
        """
        Get truststore location

        Returns:
             truststore location
        """
        t_certificate = Path(self.T_CERTIFICATE)
        if t_certificate.exists():
            return str(t_certificate)
        else:
            username = os.environ[self.HADOOP_USER_NAME]
            material_directory = Path(os.environ[self.MATERIAL_DIRECTORY])
            return str(material_directory.joinpath(username + self.TRUSTSTORE_SUFFIX))

    def _get_jks_key_store_path(self):
        """
        Get keystore location

        Returns:
             keystore location
        """
        k_certificate = Path(self.K_CERTIFICATE)
        if k_certificate.exists():
            return str(k_certificate)
        else:
            username = os.environ[self.HADOOP_USER_NAME]
            material_directory = Path(os.environ[self.MATERIAL_DIRECTORY])
            return str(material_directory.joinpath(username + self.KEYSTORE_SUFFIX))

    def _project_name(self):
        try:
            return os.environ[self.PROJECT_NAME]
        except KeyError:
            pass

        hops_user = self._project_user()
        hops_user_split = hops_user.split(
            "__"
        )  # project users have username project__user
        project = hops_user_split[0]
        return project

    def _project_user(self):
        try:
            hops_user = os.environ[self.HADOOP_USER_NAME]
        except KeyError:
            hops_user = os.environ[self.HDFS_USER]
        return hops_user

    def _get_cert_pw(self):
        """
        Get keystore password from local container

        Returns:
            Certificate password
        """
        pwd_path = Path(self.MATERIAL_PWD)
        if not pwd_path.exists():
            username = os.environ[self.HADOOP_USER_NAME]
            material_directory = Path(os.environ[self.MATERIAL_DIRECTORY])
            pwd_path = material_directory.joinpath(username + self.CERT_KEY_SUFFIX)

        with pwd_path.open() as f:
            return f.read()

    def replace_public_host(self, url):
        """replace hostname to public hostname set in HOPSWORKS_PUBLIC_HOST"""
        ui_url = url._replace(netloc=os.environ[self.HOPSWORKS_PUBLIC_HOST])
        return ui_url

    @property
    def host(self):
        return self._host



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from json import JSONEncoder
from hopsworks.client.exceptions import JobException
from hopsworks.git_file_status import GitFileStatus
from hopsworks import client
from urllib.parse import urljoin, urlparse


class Encoder(JSONEncoder):
    def default(self, obj):
        try:
            return obj.to_dict()
        except AttributeError:
            return super().default(obj)


def convert_to_abs(path, current_proj_name):
    abs_project_prefix = "/Projects/{}".format(current_proj_name)
    if not path.startswith(abs_project_prefix):
        return abs_project_prefix + "/" + path
    else:
        return path


def validate_job_conf(config, project_name):
    # User is required to set the appPath programmatically after getting the configuration
    if (
        config["type"] != "dockerJobConfiguration"
        and config["type"] != "flinkJobConfiguration"
        and "appPath" not in config
    ):
        raise JobException("'appPath' not set in job configuration")
    elif "appPath" in config and not config["appPath"].startswith("hdfs://"):
        config["appPath"] = "hdfs://" + convert_to_abs(config["appPath"], project_name)

    # If PYSPARK application set the mainClass, if SPARK validate there is a mainClass set
    if config["type"] == "sparkJobConfiguration":
        if config["appPath"].endswith(".py"):
            config["mainClass"] = "org.apache.spark.deploy.PythonRunner"
        elif "mainClass" not in config:
            raise JobException("'mainClass' not set in job configuration")

    return config


def convert_git_status_to_files(files):
    # Convert GitFileStatus to list of file paths
    if isinstance(files[0], GitFileStatus):
        tmp_files = []
        for file_status in files:
            tmp_files.append(file_status.file)
        files = tmp_files

    return files


def get_hostname_replaced_url(sub_path: str):
    """
    construct and return an url with public hopsworks hostname and sub path
    :param self:
    :param sub_path: url sub-path after base url
    :return: href url
    """
    href = urljoin(client.get_instance()._base_url, sub_path)
    url_parsed = client.get_instance().replace_public_host(urlparse(href))
    return url_parsed.geturl()



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import humps
import json
from hopsworks.core import git_provider_api
from hopsworks import util


class GitProvider:
    def __init__(
        self,
        username=None,
        token=None,
        git_provider=None,
        url=None,
        name=None,
        type=None,
        href=None,
        expand=None,
        items=None,
        count=None,
        project_id=None,
        project_name=None,
        **kwargs,
    ):
        self._username = username
        self._git_provider = git_provider

        self._git_provider_api = git_provider_api.GitProviderApi(
            project_id, project_name
        )

    @classmethod
    def from_response_json(cls, json_dict, project_id, project_name):
        # Count is not set by the backend so parse based on items array
        json_decamelized = humps.decamelize(json_dict)
        if len(json_decamelized["items"]) == 0:
            return []
        else:
            return [
                cls(**provider, project_id=project_id, project_name=project_name)
                for provider in json_decamelized["items"]
            ]

    @property
    def username(self):
        """Username set for the provider"""
        return self._username

    @property
    def git_provider(self):
        """Name of the provider, can be GitHub, GitLab or BitBucket"""
        return self._git_provider

    def delete(self):
        """Remove the git provider configuration.

        # Raises
            `RestAPIError`.
        """
        self._git_provider_api._delete_provider(self.git_provider)

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def __str__(self):
        return self.json()

    def __repr__(self):
        return f"GitProvider({self._username!r}, {self._git_provider!r})"



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import json
import humps

from hopsworks import util


class User:
    def __init__(
        self,
        username=None,
        email=None,
        first_name=None,
        firstname=None,
        last_name=None,
        lastname=None,
        status=None,
        secret=None,
        chosen_password=None,
        repeated_password=None,
        tos=None,
        two_factor=None,
        tours_state=None,
        max_num_projects=None,
        num_created_projects=None,
        test_user=None,
        user_account_type=None,
        num_active_projects=None,
        num_remaining_projects=None,
        href=None,
        **kwargs,
    ):
        self._username = username
        self._email = email

        # This is because we return two different UserDTOs with either the field firstName or firstname
        if first_name is None:
            self._first_name = firstname
        else:
            self._first_name = first_name

        if last_name is None:
            self._last_name = lastname
        else:
            self._last_name = last_name

    @classmethod
    def from_response_json(cls, json_dict):
        if json_dict:
            json_decamelized = humps.decamelize(json_dict)
            return cls(**json_decamelized)
        else:
            return None

    @property
    def username(self):
        """Username of the user"""
        return self._username

    @property
    def email(self):
        """Email of the user"""
        return self._email

    @property
    def first_name(self):
        """First name of the user"""
        return self._first_name

    @property
    def last_name(self):
        """Last name of the user"""
        return self._last_name

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def __str__(self):
        return self.json()

    def __repr__(self):
        return f"User({self._username!r}, {self._email!r}, {self._first_name!r}, {self._last_name!r})"



#
#   Copyright 2022 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import humps

from hopsworks import command


class Library:
    def __init__(
        self,
        channel,
        package_source,
        library,
        version,
        latest_version=None,
        status=None,
        preinstalled=None,
        commands=None,
        type=None,
        href=None,
        environment=None,
        project_id=None,
        **kwargs,
    ):
        self._channel = channel
        self._package_source = package_source
        self._library = library
        self._version = version
        self._latest_version = latest_version
        self._status = status
        self._preinstalled = preinstalled
        self._commands = (
            command.Command.from_response_json(commands) if commands else None
        )

        self._environment = environment
        self._project_id = project_id

    @classmethod
    def from_response_json(cls, json_dict, environment, project_id):
        json_decamelized = humps.decamelize(json_dict)
        return cls(**json_decamelized, environment=environment, project_id=project_id)



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#


class RestAPIError(Exception):
    """REST Exception encapsulating the response object and url."""

    def __init__(self, url, response):
        try:
            error_object = response.json()
        except Exception:
            error_object = {}
        message = (
            "Metadata operation error: (url: {}). Server response: \n"
            "HTTP code: {}, HTTP reason: {}, body: {}, error code: {}, error msg: {}, user "
            "msg: {}".format(
                url,
                response.status_code,
                response.reason,
                response.content,
                error_object.get("errorCode", ""),
                error_object.get("errorMsg", ""),
                error_object.get("usrMsg", ""),
            )
        )
        super().__init__(message)
        self.url = url
        self.response = response


class UnknownSecretStorageError(Exception):
    """This exception will be raised if an unused secrets storage is passed as a parameter."""


class GitException(Exception):
    """Generic git exception"""


class JobException(Exception):
    """Generic job exception"""


class EnvironmentException(Exception):
    """Generic python environment exception"""


class KafkaException(Exception):
    """Generic kafka exception"""


class DatasetException(Exception):
    """Generic dataset exception"""


class ProjectException(Exception):
    """Generic project exception"""


class OpenSearchException(Exception):
    """Generic opensearch exception"""


class JobExecutionException(Exception):
    """Generic job executions exception"""


class ExternalClientError(TypeError):
    """Raised when external client cannot be initialized due to missing arguments."""

    def __init__(self, message):
        super().__init__(message)



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import os
import furl
from abc import ABC, abstractmethod

import requests
import urllib3

from hopsworks.client import exceptions, auth
from hopsworks.decorators import connected


urllib3.disable_warnings(urllib3.exceptions.SecurityWarning)
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class Client(ABC):
    TOKEN_FILE = "token.jwt"
    APIKEY_FILE = "api.key"
    REST_ENDPOINT = "REST_ENDPOINT"
    HOPSWORKS_PUBLIC_HOST = "HOPSWORKS_PUBLIC_HOST"

    @abstractmethod
    def __init__(self):
        """To be implemented by clients."""
        pass

    def _get_verify(self, verify, trust_store_path):
        """Get verification method for sending HTTP requests to Hopsworks.

        Credit to https://gist.github.com/gdamjan/55a8b9eec6cf7b771f92021d93b87b2c

        :param verify: perform hostname verification, 'true' or 'false'
        :type verify: str
        :param trust_store_path: path of the truststore locally if it was uploaded manually to
            the external environment such as AWS Sagemaker
        :type trust_store_path: str
        :return: if verify is true and the truststore is provided, then return the trust store location
                 if verify is true but the truststore wasn't provided, then return true
                 if verify is false, then return false
        :rtype: str or boolean
        """
        if verify == "true":
            if trust_store_path is not None:
                return trust_store_path
            else:
                return True

        return False

    def _get_host_port_pair(self):
        """
        Removes "http or https" from the rest endpoint and returns a list
        [endpoint, port], where endpoint is on the format /path.. without http://

        :return: a list [endpoint, port]
        :rtype: list
        """
        endpoint = self._base_url
        if "http" in endpoint:
            last_index = endpoint.rfind("/")
            endpoint = endpoint[last_index + 1 :]
        host, port = endpoint.split(":")
        return host, port

    def _read_jwt(self):
        """Retrieve jwt from local container."""
        return self._read_file(self.TOKEN_FILE)

    def _read_apikey(self):
        """Retrieve apikey from local container."""
        return self._read_file(self.APIKEY_FILE)

    def _read_file(self, secret_file):
        """Retrieve secret from local container."""
        with open(os.path.join(self._secrets_dir, secret_file), "r") as secret:
            return secret.read()

    def _get_credentials(self, project_id):
        """Makes a REST call to hopsworks for getting the project user certificates needed to connect to services such as Hive

        :param project_id: id of the project
        :type project_id: int
        :return: JSON response with credentials
        :rtype: dict
        """
        return self._send_request("GET", ["project", project_id, "credentials"])

    def _write_pem_file(self, content: str, path: str) -> None:
        with open(path, "w") as f:
            f.write(content)

    @connected
    def _send_request(
        self,
        method,
        path_params,
        query_params=None,
        headers=None,
        data=None,
        stream=False,
        files=None,
        with_base_path_params=True,
    ):
        """Send REST request to Hopsworks.

        Uses the client it is executed from. Path parameters are url encoded automatically.

        :param method: 'GET', 'PUT' or 'POST'
        :type method: str
        :param path_params: a list of path params to build the query url from starting after
            the api resource, for example `["project", 119, "featurestores", 67]`.
        :type path_params: list
        :param query_params: A dictionary of key/value pairs to be added as query parameters,
            defaults to None
        :type query_params: dict, optional
        :param headers: Additional header information, defaults to None
        :type headers: dict, optional
        :param data: The payload as a python dictionary to be sent as json, defaults to None
        :type data: dict, optional
        :param stream: Set if response should be a stream, defaults to False
        :type stream: boolean, optional
        :param files: dictionary for multipart encoding upload
        :type files: dict, optional
        :raises RestAPIError: Raised when request wasn't correctly received, understood or accepted
        :return: Response json
        :rtype: dict
        """
        f_url = furl.furl(self._base_url)
        if with_base_path_params:
            base_path_params = ["hopsworks-api", "api"]
            f_url.path.segments = base_path_params + path_params
        else:
            f_url.path.segments = path_params
        url = str(f_url)

        request = requests.Request(
            method,
            url=url,
            headers=headers,
            data=data,
            params=query_params,
            auth=self._auth,
            files=files,
        )

        prepped = self._session.prepare_request(request)
        response = self._session.send(prepped, verify=self._verify, stream=stream)

        if response.status_code == 401 and self.REST_ENDPOINT in os.environ:
            # refresh token and retry request - only on hopsworks
            self._auth = auth.BearerAuth(self._read_jwt())
            # Update request with the new token
            request.auth = self._auth
            prepped = self._session.prepare_request(request)
            response = self._session.send(prepped, verify=self._verify, stream=stream)

        if response.status_code // 100 != 2:
            raise exceptions.RestAPIError(url, response)

        if stream:
            return response
        else:
            # handle different success response codes
            if len(response.content) == 0:
                return None
            return response.json()

    def _close(self):
        """Closes a client. Can be implemented for clean up purposes, not mandatory."""
        self._connected = False



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import humps
import json
from hopsworks.core import kafka_api
from hopsworks import util


class KafkaTopic:
    def __init__(
        self,
        name=None,
        num_of_replicas=None,
        num_of_partitions=None,
        schema_name=None,
        schema_version=None,
        schema_content=None,
        owner_project_id=None,
        shared=None,
        accepted=None,
        project_id=None,
        project_name=None,
        type=None,
        href=None,
        expand=None,
        items=None,
        count=None,
        **kwargs,
    ):
        self._name = name
        self._num_of_replicas = num_of_replicas
        self._num_of_partitions = num_of_partitions
        self._schema_name = schema_name
        self._schema_version = schema_version
        self._schema_content = schema_content
        self._owner_project_id = owner_project_id
        self._shared = shared
        self._accepted = accepted

        self._kafka_api = kafka_api.KafkaApi(project_id, project_name)

    @classmethod
    def from_response_json(cls, json_dict, project_id, project_name):
        json_decamelized = humps.decamelize(json_dict)
        if "count" not in json_decamelized:
            return cls(
                **json_decamelized, project_id=project_id, project_name=project_name
            )
        elif json_decamelized["count"] == 0:
            return []
        else:
            return [
                cls(**kafka_topic, project_id=project_id, project_name=project_name)
                for kafka_topic in json_decamelized["items"]
            ]

    def update_from_response_json(self, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        self.__init__(**json_decamelized)
        return self

    @property
    def name(self):
        """Name of the topic"""
        return self._name

    @property
    def replicas(self):
        """Replication factor for the topic"""
        return self._num_of_replicas

    @property
    def partitions(self):
        """Number of partitions for the topic"""
        return self._num_of_partitions

    @property
    def schema(self):
        """Schema for the topic"""
        return self._kafka_api._get_schema_details(
            self._schema_name, self._schema_version
        )

    def delete(self):
        """Delete the topic
        !!! danger "Potentially dangerous operation"
            This operation deletes the topic.
        # Raises
            `RestAPIError`.
        """
        self._kafka_api._delete_topic(self.name)

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def __str__(self):
        return self.json()

    def __repr__(self):
        return f"KafkaTopic({self._name!r})"



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from hopsworks import client, execution


class ExecutionsApi:
    def __init__(
        self,
        project_id,
    ):
        self._project_id = project_id

    def _start(self, job, args: str = None):
        _client = client.get_instance()
        path_params = ["project", self._project_id, "jobs", job.name, "executions"]

        return execution.Execution.from_response_json(
            _client._send_request("POST", path_params, data=args), self._project_id, job
        )

    def _get(self, job, id):
        _client = client.get_instance()
        path_params = [
            "project",
            self._project_id,
            "jobs",
            job.name,
            "executions",
            id,
        ]

        headers = {"content-type": "application/json"}
        return execution.Execution.from_response_json(
            _client._send_request("GET", path_params, headers=headers),
            self._project_id,
            job,
        )

    def _get_all(self, job):
        _client = client.get_instance()
        path_params = ["project", self._project_id, "jobs", job.name, "executions"]

        query_params = {"sort_by": "submissiontime:desc"}

        headers = {"content-type": "application/json"}
        return execution.Execution.from_response_json(
            _client._send_request(
                "GET", path_params, headers=headers, query_params=query_params
            ),
            self._project_id,
            job,
        )

    def _delete(self, job_name, id):
        _client = client.get_instance()
        path_params = [
            "project",
            self._project_id,
            "jobs",
            job_name,
            "executions",
            id,
        ]
        _client._send_request("DELETE", path_params)



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from hopsworks import client, kafka_topic, kafka_schema, constants
from hopsworks.client.exceptions import KafkaException
import json
import socket
from hopsworks.client.external import Client


class KafkaApi:
    def __init__(
        self,
        project_id,
        project_name,
    ):
        self._project_id = project_id
        self._project_name = project_name

    def create_topic(
        self,
        name: str,
        schema: str,
        schema_version: int,
        replicas: int = 1,
        partitions: int = 1,
    ):
        """Create a new kafka topic.

        ```python

        import hopsworks

        project = hopsworks.login()

        kafka_api = project.get_kafka_api()

        kafka_topic = kafka_api.create_topic("my_topic", "my_schema", 1)

        ```
        # Arguments
            name: name of the topic
            schema: subject name of the schema
            schema_version: version of the schema
            replicas: replication factor for the topic
            partitions: partitions for the topic
        # Returns
            `KafkaTopic`: The KafkaTopic object
        # Raises
            `RestAPIError`: If unable to create the topic
        """
        _client = client.get_instance()

        path_params = ["project", self._project_id, "kafka", "topics"]
        data = {
            "name": name,
            "schemaName": schema,
            "schemaVersion": schema_version,
            "numOfReplicas": replicas,
            "numOfPartitions": partitions,
        }

        headers = {"content-type": "application/json"}
        return kafka_topic.KafkaTopic.from_response_json(
            _client._send_request(
                "POST", path_params, headers=headers, data=json.dumps(data)
            ),
            self._project_id,
            self._project_name,
        )

    def create_schema(self, subject: str, schema: dict):
        """Create a new kafka schema.

        ```python

        import hopsworks

        project = hopsworks.login()

        kafka_api = project.get_kafka_api()

        avro_schema = {
          "type": "record",
          "name": "tutorial",
          "fields": [
            {
              "name": "id",
              "type": "int"
            },
            {
              "name": "data",
              "type": "string"
            }
          ]
        }

        kafka_topic = kafka_api.create_schema("my_schema", avro_schema)

        ```
        # Arguments
            subject: subject name of the schema
            schema: avro schema definition
        # Returns
            `KafkaSchema`: The KafkaSchema object
        # Raises
            `RestAPIError`: If unable to create the schema
        """
        _client = client.get_instance()

        path_params = [
            "project",
            self._project_id,
            "kafka",
            "subjects",
            subject,
            "versions",
        ]

        headers = {"content-type": "application/json"}
        schema = kafka_schema.KafkaSchema.from_response_json(
            _client._send_request(
                "POST",
                path_params,
                headers=headers,
                data=json.dumps({"schema": json.dumps(schema)}),
            ),
            self._project_id,
            self._project_name,
        )
        # TODO: Fix backend, GET request required as POST does not set schema field in the returned payload
        return self.get_schema(schema.subject, schema.version)

    def get_topic(self, name: str):
        """Get kafka topic by name.

        # Arguments
            name: name of the topic
        # Returns
            `KafkaTopic`: The KafkaTopic object
        # Raises
            `RestAPIError`: If unable to get the topic
        """
        topics = self.get_topics()

        for topic in topics:
            if topic.name == name:
                return topic

        raise KafkaException("No topic named {} could be found".format(name))

    def get_topics(self):
        """Get all kafka topics.

        # Returns
            `List[KafkaTopic]`: List of KafkaTopic objects
        # Raises
            `RestAPIError`: If unable to get the topics
        """
        _client = client.get_instance()
        path_params = ["project", self._project_id, "kafka", "topics"]

        return kafka_topic.KafkaTopic.from_response_json(
            _client._send_request("GET", path_params),
            self._project_id,
            self._project_name,
        )

    def _delete_topic(self, name: str):
        """Delete the topic.
        :param name: name of the topic
        :type name: str
        """
        _client = client.get_instance()
        path_params = [
            "project",
            self._project_id,
            "kafka",
            "topics",
            name,
        ]
        _client._send_request("DELETE", path_params)

    def _delete_subject_version(self, subject: str, version: int):
        """Delete the schema.
        :param subject: subject name of the schema
        :type subject: str
        :param version: version of the subject
        :type version: int
        """
        _client = client.get_instance()
        path_params = [
            "project",
            self._project_id,
            "kafka",
            "subjects",
            subject,
            "versions",
            str(version),
        ]
        _client._send_request("DELETE", path_params)

    def get_subjects(self):
        """Get all subjects.

        # Returns
            `List[str]`: List of registered subjects
        # Raises
            `RestAPIError`: If unable to get the subjects
        """
        topics = self.get_topics()

        subjects = set()

        for topic in topics:
            subjects.add(topic.schema.subject)

        return list(subjects)

    def get_schemas(self, subject: str):
        """Get all schema versions for the subject.

        # Arguments
            subject: subject name
        # Returns
            `List[KafkaSchema]`: List of KafkaSchema objects
        # Raises
            `RestAPIError`: If unable to get the schemas
        """
        _client = client.get_instance()
        path_params = [
            "project",
            self._project_id,
            "kafka",
            "subjects",
            subject,
            "versions",
        ]

        versions = _client._send_request("GET", path_params)

        schemas = []
        for version in versions:
            schemas.append(self._get_schema_details(subject, version))

        return schemas

    def get_schema(self, subject: str, version: int):
        """Get schema given subject name and version.

        # Arguments
            subject: subject name
            version: version number
        # Returns
            `KafkaSchema`: KafkaSchema object
        # Raises
            `RestAPIError`: If unable to get the schema
        """
        schemas = self.get_schemas(subject)
        for schema in schemas:
            if schema.version == version:
                return schema

        raise KafkaException(
            "No schema for subject {} and version {} could be found".format(
                subject, version
            )
        )

    def _get_schema_details(self, subject: str, version: int):
        """Get the schema details.
        :param subject: subject name of the schema
        :type subject: str
        :param version: version of the subject
        :type version: int
        """
        _client = client.get_instance()
        path_params = [
            "project",
            self._project_id,
            "kafka",
            "subjects",
            subject,
            "versions",
            str(version),
        ]

        return kafka_schema.KafkaSchema.from_response_json(
            _client._send_request("GET", path_params),
            self._project_id,
            self._project_name,
        )

    def _get_broker_endpoints(self, externalListeners: bool = False):
        _client = client.get_instance()
        path_params = [
            "project",
            self._project_id,
            "kafka",
            "clusterinfo",
        ]
        query_params = {"external": externalListeners}
        headers = {"content-type": "application/json"}
        return _client._send_request(
            "GET", path_params, query_params=query_params, headers=headers
        )["brokers"]

    def _get_security_protocol(self):
        """
        Gets the security protocol used for communicating with Kafka brokers in a Hopsworks cluster
        Returns:
            the security protocol for communicating with Kafka brokers in a Hopsworks cluster
        """
        return constants.KAFKA_SSL_CONFIG.SSL

    def get_default_config(self):
        """Get the configuration to set up a Producer or Consumer for a Kafka broker using confluent-kafka.

        ```python

        import hopsworks

        project = hopsworks.login()

        kafka_api = project.get_kafka_api()

        kafka_conf = kafka_api.get_default_config()

        from confluent_kafka import Producer

        producer = Producer(kafka_conf)

        ```
        # Returns
            `dict`: The kafka configuration
        # Raises
            `RestAPIError`: If unable to get the kafka configuration.
        """

        config = {
            constants.KAFKA_SSL_CONFIG.SECURITY_PROTOCOL_CONFIG: self._get_security_protocol(),
            constants.KAFKA_SSL_CONFIG.SSL_CA_LOCATION_CONFIG: client.get_instance()._get_ca_chain_path(
                self._project_name
            ),
            constants.KAFKA_SSL_CONFIG.SSL_CERTIFICATE_LOCATION_CONFIG: client.get_instance()._get_client_cert_path(
                self._project_name
            ),
            constants.KAFKA_SSL_CONFIG.SSL_PRIVATE_KEY_LOCATION_CONFIG: client.get_instance()._get_client_key_path(
                self._project_name
            ),
            constants.KAFKA_CONSUMER_CONFIG.CLIENT_ID_CONFIG: socket.gethostname(),
            constants.KAFKA_CONSUMER_CONFIG.GROUP_ID_CONFIG: "my-group-id",
            constants.KAFKA_SSL_CONFIG.SSL_ENDPOINT_IDENTIFICATION_ALGORITHM_CONFIG: "none",
        }
        _client = client.get_instance()
        if type(_client) == Client:
            config[constants.KAFKA_PRODUCER_CONFIG.BOOTSTRAP_SERVERS_CONFIG] = ",".join(
                [
                    endpoint.replace("EXTERNAL://", "")
                    for endpoint in self._get_broker_endpoints(externalListeners=True)
                ]
            )
        else:
            config[constants.KAFKA_PRODUCER_CONFIG.BOOTSTRAP_SERVERS_CONFIG] = ",".join(
                [
                    endpoint.replace("INTERNAL://", "")
                    for endpoint in self._get_broker_endpoints(externalListeners=False)
                ]
            )

        return config



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import warnings
import logging
import os
import sys
import getpass
import tempfile
from pathlib import Path

from hopsworks.client.exceptions import RestAPIError, ProjectException
from hopsworks import version, constants, client
from hopsworks.connection import Connection

# Needs to run before import of hsml and hsfs
warnings.filterwarnings(action="ignore", category=UserWarning, module=r".*psycopg2")

import hsml  # noqa: F401, E402
import hsfs  # noqa: F401, E402

__version__ = version.__version__

connection = Connection.connection

_hw_connection = Connection.connection

_connected_project = None


def hw_formatwarning(message, category, filename, lineno, line=None):
    return "{}: {}\n".format(category.__name__, message)


warnings.formatwarning = hw_formatwarning

__all__ = ["connection"]

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)s: %(message)s",
    stream=sys.stdout,
)


def login(
    host: str = None,
    port: int = 443,
    project: str = None,
    api_key_value: str = None,
    api_key_file: str = None,
):
    """Connect to [Serverless Hopsworks](https://app.hopsworks.ai) by calling the `hopsworks.login()` function with no arguments.

    ```python

    project = hopsworks.login()

    ```

    Alternatively, connect to your own Hopsworks installation by specifying the host, port and api key.

    ```python

    project = hopsworks.login(host="my.hopsworks.server",
                              port=8181,
                              api_key_value="DKN8DndwaAjdf98FFNSxwdVKx")

    ```

    In addition to setting function arguments directly, `hopsworks.login()` also reads the environment variables:
    HOPSWORKS_HOST, HOPSWORKS_PORT, HOPSWORKS_PROJECT and HOPSWORKS_API_KEY.

    The function arguments do however take precedence over the environment variables in case both are set.

    # Arguments
        host: The hostname of the Hopsworks instance, defaults to `None`.
        port: The port on which the Hopsworks instance can be reached,
            defaults to `443`.
        project: Name of the project to access. If used inside a Hopsworks environment it always gets the current project. If not provided you will be prompted to enter it.
        api_key_value: Value of the Api Key
        api_key_file: Path to file wih Api Key
    # Returns
        `Project`: The Project object
    # Raises
        `RestAPIError`: If unable to connect to Hopsworks
    """

    global _connected_project

    # If already logged in, should reset connection and follow login procedure as Connection may no longer be valid
    logout()

    global _hw_connection

    # If inside hopsworks, just return the current project for now
    if "REST_ENDPOINT" in os.environ:
        _hw_connection = _hw_connection()
        _connected_project = _hw_connection.get_project()
        print("\nLogged in to project, explore it here " + _connected_project.get_url())
        return _connected_project

    # This is run for an external client

    # Function arguments takes precedence over environment variable.
    # Here we check if environment variable exists and function argument is not set, we use the environment variable.

    # If api_key_value/api_key_file not defined, get HOPSWORKS_API_KEY environment variable
    api_key = None
    if (
        api_key_value is None
        and api_key_file is None
        and "HOPSWORKS_API_KEY" in os.environ
    ):
        api_key = os.environ["HOPSWORKS_API_KEY"]

    # If project argument not defined, get HOPSWORKS_PROJECT environment variable
    if project is None and "HOPSWORKS_PROJECT" in os.environ:
        project = os.environ["HOPSWORKS_PROJECT"]

    # If host argument not defined, get HOPSWORKS_HOST environment variable
    if host is None and "HOPSWORKS_HOST" in os.environ:
        host = os.environ["HOPSWORKS_HOST"]
    elif host is None:  # Always do a fallback to Serverless Hopsworks if not defined
        host = constants.HOSTS.APP_HOST

    # If port same as default, get HOPSWORKS_HOST environment variable
    if port == 443 and "HOPSWORKS_PORT" in os.environ:
        port = os.environ["HOPSWORKS_PORT"]

    # This .hw_api_key is created when a user logs into Serverless Hopsworks the first time.
    # It is then used only for future login calls to Serverless. For other Hopsworks installations it's ignored.
    api_key_path = _get_cached_api_key_path()

    # Conditions for getting the api_key
    # If user supplied the api key directly
    if api_key_value is not None:
        api_key = api_key_value
    # If user supplied the api key in a file
    elif api_key_file is not None:
        file = None
        if os.path.exists(api_key_file):
            try:
                file = open(api_key_file, mode="r")
                api_key = file.read()
            finally:
                file.close()
        else:
            raise IOError(
                "Could not find api key file on path: {}".format(api_key_file)
            )
    # If user connected to Serverless Hopsworks, and the cached .hw_api_key exists, then use it.
    elif os.path.exists(api_key_path) and host == constants.HOSTS.APP_HOST:
        try:
            _hw_connection = _hw_connection(
                host=host, port=port, api_key_file=api_key_path
            )
            _connected_project = _prompt_project(_hw_connection, project)
            print(
                "\nLogged in to project, explore it here "
                + _connected_project.get_url()
            )
            return _connected_project
        except RestAPIError:
            logout()
            # API Key may be invalid, have the user supply it again
            os.remove(api_key_path)

    if api_key is None and host == constants.HOSTS.APP_HOST:
        print(
            "Copy your Api Key (first register/login): https://c.app.hopsworks.ai/account/api/generated"
        )
        api_key = getpass.getpass(prompt="\nPaste it here: ")

        # If api key was provided as input, save the API key locally on disk to avoid users having to enter it again in the same environment
        Path(os.path.dirname(api_key_path)).mkdir(parents=True, exist_ok=True)
        descriptor = os.open(
            path=api_key_path, flags=(os.O_WRONLY | os.O_CREAT | os.O_TRUNC), mode=0o600
        )
        with open(descriptor, "w") as fh:
            fh.write(api_key.strip())

    try:
        _hw_connection = _hw_connection(host=host, port=port, api_key_value=api_key)
        _connected_project = _prompt_project(_hw_connection, project)
    except RestAPIError as e:
        logout()
        raise e

    print("\nLogged in to project, explore it here " + _connected_project.get_url())
    return _connected_project


def _get_cached_api_key_path():
    """
    This function is used to get an appropriate path to store the user supplied API Key for Serverless Hopsworks.

    First it will search for .hw_api_key in the current working directory, if it exists it will use it (this is default in 3.0 client)
    Otherwise, falls back to storing the API key in HOME
    If not sufficient permissions are set in HOME to create the API key (writable and executable), it uses the temp directory to store it.

    """

    api_key_name = ".hw_api_key"
    api_key_folder = ".{}_hopsworks_app".format(getpass.getuser())

    # Path for current working directory api key
    cwd_api_key_path = os.path.join(os.getcwd(), api_key_name)

    # Path for home api key
    home_dir_path = os.path.expanduser("~")
    home_api_key_path = os.path.join(home_dir_path, api_key_folder, api_key_name)

    # Path for tmp api key
    temp_dir_path = tempfile.gettempdir()
    temp_api_key_path = os.path.join(temp_dir_path, api_key_folder, api_key_name)

    if os.path.exists(
        cwd_api_key_path
    ):  # For backward compatibility, if .hw_api_key exists in current working directory get it from there
        api_key_path = cwd_api_key_path
    elif os.access(home_dir_path, os.W_OK) and os.access(
        home_dir_path, os.X_OK
    ):  # Otherwise use the home directory of the user
        api_key_path = home_api_key_path
    else:  # Finally try to store it in temp
        api_key_path = temp_api_key_path

    return api_key_path


def _prompt_project(valid_connection, project):
    saas_projects = valid_connection.get_projects()
    if project is None:
        if len(saas_projects) == 0:
            raise ProjectException("Could not find any project")
        elif len(saas_projects) == 1:
            return saas_projects[0]
        else:
            while True:
                print("\nMultiple projects found. \n")
                for index in range(len(saas_projects)):
                    print("\t (" + str(index + 1) + ") " + saas_projects[index].name)
                while True:
                    project_index = input("\nEnter project to access: ")
                    # Handle invalid input type
                    try:
                        project_index = int(project_index)
                        # Handle negative indexing
                        if project_index <= 0:
                            print("Invalid input, must be greater than or equal to 1")
                            continue
                        # Handle index out of range
                        try:
                            return saas_projects[project_index - 1]
                        except IndexError:
                            print(
                                "Invalid input, should be an integer from the list of projects."
                            )
                    except ValueError:
                        print(
                            "Invalid input, should be an integer from the list of projects."
                        )
    else:
        for proj in saas_projects:
            if proj.name == project:
                return proj
        raise ProjectException("Could not find project {}".format(project))


def logout():
    global _hw_connection
    if isinstance(_hw_connection, Connection):
        _hw_connection.close()
    client.stop()
    _hw_connection = Connection.connection
